package jmaki.runtime.config;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import jmaki.runtime.ResourceManager;
import jmaki.runtime.AjaxGlobalObjects;
import jmaki.runtime.AjaxException;
import jmaki.runtime.IOUtil;

public class GlobalConfig {
    
    private static final String GLOBALCONFIG="/config.json";
    private static final String WIDGETCONFIG="/widget.json";
    private static final String GLOBALBOOTSTRAPSCRIPT = "/jmaki.js";
    
    private HashMap globalTypes = null; // maybe null if all config-elements are in widget.json
    
    private ApiKeyMap globalApiKeys = null; // maybe null if all config-elements are in widget.json
    
    private HashMap widgetConfigs = new HashMap(); // without widgetConfigs there ain't no widgets
    
    private LinkedHashSet glueIncludes = null;
    private ResourceManager resourceManager;
    
    public static final Iterator NULLITERATOR = (Iterator) new Iterator() {
        public Object next() throws NoSuchElementException { throw new NoSuchElementException(); }
        public boolean hasNext() { return false; }
        public void remove() throws IllegalStateException { throw new IllegalStateException("next has not been called yet"); }
    };
    
     /**
      * <code>true/code> if the config.json for this application is to be
      * loaded from the web app resources.  <code>false</code> if it's to be
      * loaded from the classpath.
      */
    private boolean isGlobalConfigWebAppResource = false;

    /**
     * <code>true/code> if the jmaki.js for this application is to be
     * loaded from the web app resources.  <code>false</code> if it's to be
     * loaded from the classpath.
     */
    private boolean isGlobalBootstrapScriptWebAppResource = false;
     
    public Iterator getGlueIncludes() {
        return glueIncludes == null ? NULLITERATOR : glueIncludes.iterator();
    }
    
    public boolean isGlobalConfigWebAppResource() {	return this.isGlobalConfigWebAppResource; }
    
    public boolean isGlobalBootstrapScriptWebAppResource() { return this.isGlobalBootstrapScriptWebAppResource; }
    
    public GlobalConfig(ResourceManager resourceManager) throws AjaxException {
        this.resourceManager = resourceManager;
        if (resourceManager.isClassPathResource(resourceManager.getWebResourcePath() + GLOBALCONFIG)) {
            loadGlobalConfig(resourceManager.locateClassPathResource(resourceManager.getWebResourcePath() + GLOBALCONFIG));
        }
        if (resourceManager.isWebAppResource(resourceManager.getWebResourcePath() + GLOBALCONFIG)) {
            loadGlobalConfig(resourceManager.locateWebAppResource(resourceManager.getWebResourcePath() + GLOBALCONFIG));
            isGlobalConfigWebAppResource = true;
        }
        if (resourceManager.isWebAppResource(resourceManager.getResourcesDir() + GLOBALBOOTSTRAPSCRIPT)) {
            isGlobalBootstrapScriptWebAppResource = true;
    	}
    }
    
    public Config getConfig(String widgetname, String location) throws AjaxException {
        WidgetConfig widgetConfig = null;
        Type type = null;
        if (location != null && (widgetConfig = (WidgetConfig)widgetConfigs.get(location))==null) {
            widgetConfig = loadWidgetConfig(widgetname, location);
            widgetConfigs.put(location,widgetConfig);
        }
        int positionOfDot;
        if ((type=widgetConfig.getType())==null && globalTypes!=null) {
        	if ((type=(Type)globalTypes.get(widgetname))==null && ((positionOfDot=widgetname.indexOf('.'))<1))
        		type = (Type)globalTypes.get(widgetname.substring(0,positionOfDot));
        }
        return new ConfigImpl(type,globalApiKeys,widgetConfig.getApiKeyMap());
    }
    
    public void discard() {
        if (this.globalApiKeys!=null) {
            this.globalApiKeys.discard();
            this.globalApiKeys=null;
        }
        if (this.globalTypes!=null) {
            Iterator globalTypesIterator = this.globalTypes.entrySet().iterator();
            while (globalTypesIterator.hasNext()) {
                ((Type)((Map.Entry)globalTypesIterator.next()).getValue()).discard();
            }
            this.globalTypes.clear();
            this.globalTypes=null;
        }
        if (this.widgetConfigs!=null) {
            Iterator widgetConfigsIterator = widgetConfigs.entrySet().iterator();
            while (widgetConfigsIterator.hasNext()) {
                ((WidgetConfig)((Map.Entry)widgetConfigsIterator.next()).getValue()).discard();
            }
            this.widgetConfigs.clear();
            this.widgetConfigs=null;
        }
    }
    
    private synchronized void loadGlobalConfig(URL url) throws AjaxException {
        InputStream is = null;
        try {           
            is = url.openConnection().getInputStream();           
            if (is != null) {
                JSONObject jAll = IOUtil.loadJSONObject(is);               
                if (jAll.has("config")) {
                    JSONObject dConfig = jAll.getJSONObject("config");
                    if (dConfig.has("glue")) {
                        JSONObject glue = dConfig.getJSONObject("glue");
                        if (glue.has("includes")){
                        	if (glueIncludes==null)
                        		glueIncludes=new LinkedHashSet();
                        	JSONArray glueList = glue.getJSONArray("includes"); 
                            for (int il = 0; il < glueList.length(); il++) {
                                GlueInclude include = loadGlueInclude(glueList.get(il));
                                if (include!=null) glueIncludes.add(include);
                            }
                        }
                    }
                    if (dConfig.has("types")) {
                        JSONArray typesArray = dConfig.getJSONArray("types");
                        for (int l = 0; l < typesArray.length(); l++) {
                            JSONObject jType = typesArray.getJSONObject(l);
                            String keytype = jType.getString("id");
                            if (globalTypes==null)
                                globalTypes=new HashMap();
                            globalTypes.put(keytype,loadType(null,jType));
                        }
                    }
                    if (dConfig.has("apikeys")) {
                        JSONArray apikeyArray = dConfig.getJSONArray("apikeys");
                        for (int l = 0; l < apikeyArray.length(); l++) {
                            JSONObject jApiKeys = apikeyArray.getJSONObject(l);
                            String keytype = jApiKeys.getString("id");
                            if (globalApiKeys==null)
                                globalApiKeys=new ApiKeyMap();
                            List keyList = loadApiKeys(jApiKeys,(List)globalApiKeys.get(keytype));
                            if (keyList != null && !globalApiKeys.containsKey(keytype)) {
                                globalApiKeys.put(keytype, keyList);
                            }
                        }
                    }
                }
            }
        } catch (JSONException ex) {
            throw new WidgetConfigException("error parsing 'config.json' from location: '"+url+"': "+ex.getMessage(),ex);
        } catch (IOException ioe) {
            throw new WidgetConfigException("error loading 'config.json' from location: '"+url+"': "+ioe.getMessage(),ioe);
        } finally {
            if (is!=null) {
                try {
                    is.close();
                } catch (IOException ioe) {
                    throw new WidgetConfigException("error closing inputstream for url: '"+url+"'",ioe);
                }
            }
        }
    }

    private synchronized WidgetConfig loadWidgetConfig(String widgetname, String location) throws AjaxException {
        URL url = null;
        WidgetConfig widgetConfig = new WidgetConfig();
        String widgetPath = location+WIDGETCONFIG;
        if (resourceManager.isWebAppResource(widgetPath)) {
        	url = resourceManager.locateWebAppResource(widgetPath);
        } else
        	if (resourceManager.isClassPathResource(widgetPath))
        		url = resourceManager.locateClassPathResource(widgetPath);
        InputStream is = null;
        try {
            if (url != null) is = url.openConnection().getInputStream();
            if (is != null) {
                JSONObject jAll = IOUtil.loadJSONObject(is);
                if (jAll.has("config")) {
                    JSONObject dConfig = jAll.getJSONObject("config");
                    if (dConfig.has("type")) {
                        JSONObject jType = dConfig.getJSONObject("type");
                        int positionOfDot = widgetname.indexOf('.');
                        String widgettype = positionOfDot>0 ? widgetname.substring(0,positionOfDot) : widgetname;
                        String keytype = jType.getString("id");
                        // gmurray71 : This logic will not work.
                        // The type 'id' is used to represent whether or not a resource has been loaded and multiple widgets
                        // Such as the google widgets use the *same* type which means that the URLs should only be written one
                        // time to the page. There is no correlation between the type id of a widget and the widget name
                        // However if a type has been written once to the page it should not be re-written.
                        // A good test for this is the googlemap popup widget which has three of the same widgets in the same page
                        // What we don't want is for three google api links to get rendered.
                        //if (keytype.equals(widgetname) || keytype.equals(widgettype))
                            widgetConfig.setType(loadType(location,jType));
                        //else throw new WidgetConfigException("unexpected element 'id="+keytype+"' in widget.json for '"+widgetname+"'");
                    }
                    if (dConfig.has("apikeys")) {
                        JSONArray apikeyArray = dConfig.getJSONArray("apikeys");
                        if (apikeyArray.length()>0) {
                        	ApiKeyMap apiKeys = new ApiKeyMap();
                        	for (int l = 0; l < apikeyArray.length(); l++) {
                        		JSONObject jApiKeys = apikeyArray.getJSONObject(l);
                        		String keytype = jApiKeys.getString("id");
                        		List keyList = loadApiKeys(jApiKeys,(List)apiKeys.get(keytype));
                        		if (keyList != null && !apiKeys.containsKey(keytype)) {
                        			apiKeys.put(keytype, keyList);
                        		}
                        	}
                        	widgetConfig.setApiKeys(apiKeys);
                        }
                    }
                }
            }
        } catch (JSONException ex) {
            throw new WidgetConfigException("error parsing 'widget.json' from location: '"+url+"': "+ex.getMessage(),ex);
        } catch (IOException ioe) {
            throw new WidgetConfigException("error loading 'widget.json' from location: '"+url+"': "+ioe.getMessage(),ioe);
        } finally {
            if (is!=null) {
                try {
                    is.close();
                } catch (IOException ioe) {
                    throw new WidgetConfigException("error closing inputstream for url: '"+url+"'",ioe);
                }
            }
        }
        return widgetConfig;
    }
    
    /**
     *  Load a list of GlueInclude Objects which are matched against the current URL
     *  and the script included after the jmaki.js script is written.
     *
     *  GlueIncludes are loadded once for the entire web applicaton and
     *  can be described in the config.json a JSONArray of Strings or
     *  a JSONArray of JSONObjects.
     *
     *  In the case of JSONStrings the GlueInclude is mapped to all requests
     *  In the case of JSONObjects the GlueInclude is looks at the URL of the
     *  current request and will render the GlueInclude url matches the 
     *  the url (from the context root of the applicaton).
     *
     *  Since glue includes can be included from a classpath resource and the current 
     *  web app resources we want to combine the resources if both a classpath and
     *  web app config.json are provided.
     *
     *We should do an equality check to make sure duplicate glue entries
     *are not added.
     * 
     */
    private GlueInclude loadGlueInclude(Object obj) throws JSONException {
    	// this supports the old way of just a straight JSON array of strings
    	if (obj instanceof JSONObject) {
    		JSONObject jo = (JSONObject)obj;
    		return new GlueInclude(jo.getString("url"), jo.getString("lib"));
    	} else if (obj instanceof String) {
    		return new GlueInclude("*", (String)obj);
    	}
    	return null;
    }
    
    private Type loadType(String currentDir, JSONObject jType) throws JSONException, WidgetConfigException {
        Type type = new Type(jType.getString("id"));
        if (jType.has("styles")) {
            JSONArray nA = jType.getJSONArray("styles");
            for (int il = 0; il < nA.length(); il++) {
                String s = nA.getString(il);
                if (type.getStyles()==null)
                    type.setStyles(new ArrayList());
                type.getStyles().add(IOUtil.resolvePath(currentDir,s));
            }
        }
        if (jType.has("libs")) {
            JSONArray nA = jType.getJSONArray("libs");
            for (int il = 0; il < nA.length(); il++) {
                String s = nA.getString(il);
                if (type.getLibs()==null)
                    type.setLibs(new ArrayList());
                type.getLibs().add(IOUtil.resolvePath(currentDir,s));
            }
        }
        if (jType.has("preload"))
            type.setPreLoad(jType.getString("preload"));
        if (jType.has("postload"))
            type.setPostLoad (jType.getString("postload"));
        if (jType.has("apikey"))
            type.setApiKeyType(jType.getString("apikey"));
        if (jType.has("dynamicallyLoadable"))
            type.setDynamicallyLoadable(jType.getBoolean("dynamicallyLoadable"));        
        return type;
    }
    
    private List loadApiKeys(JSONObject jApiKeys, List keyList) throws JSONException {
        JSONArray keys = jApiKeys.getJSONArray("keys");
        int keyListLen = keys.length();
        if (keyListLen>0) {
            if (keyList==null)
                keyList=new ArrayList();
            for (int k = 0; k < keyListLen; k++) {
                JSONObject apiKey = keys.getJSONObject(k);
                String targeturl = apiKey.getString("url");
                if (targeturl.endsWith("/"))
                    targeturl=targeturl.substring(0, targeturl.length()-1);
                keyList.add(new ApiKey(targeturl,apiKey.getString("key")));
            }
        }
        return keyList;
    }
}
